@@ -1,6 +1,6 @@
 /* 
  * 
- * Copyright (C) 2009-2010 Colomban Wendling <ban@herbesfolles.org>
+ * Copyright (C) 2009-2011 Colomban Wendling <ban@herbesfolles.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -105,7 +105,7 @@ ctpl_eval_operator_minus (CtplValue  *lvalue,
   return rv;
 }
 
-/* Tries to evaluate a addition operation */
+/* Tries to evaluate an addition operation */
 static gboolean
 ctpl_eval_operator_plus (CtplValue *lvalue,
                          CtplValue *rvalue,
@@ -122,11 +122,12 @@ ctpl_eval_operator_plus (CtplValue *lvalue,
           const GSList *array;
           
           for (array = ctpl_value_get_array (rvalue);
-               array->data;
+               array;
                array = array->next) {
             ctpl_value_array_append (value, array->data);
           }
         }
+        break;
         
         default:
           ctpl_value_array_append (value, rvalue);
@@ -164,8 +165,8 @@ ctpl_eval_operator_plus (CtplValue *lvalue,
           gchar buf[G_ASCII_DTOSTR_BUF_SIZE];
           
           tmp = g_strconcat (ctpl_value_get_string (lvalue),
-                             g_ascii_dtostr (buf, sizeof (buf),
-                                             ctpl_value_get_float (rvalue)),
+                             ctpl_math_dtostr (buf, sizeof (buf),
+                                               ctpl_value_get_float (rvalue)),
                              NULL);
         } else if (CTPL_VALUE_HOLDS_INT (rvalue)) {
           tmp = g_strdup_printf ("%s%ld", ctpl_value_get_string (lvalue),
@@ -237,7 +238,6 @@ do_multiply_string (const gchar  *str,
       }
     }
   }
-  //~ g_debug ("mutiplied string: '%s'", buf);
   
   return buf;
 }
@@ -374,13 +374,132 @@ ctpl_eval_operator_div (CtplValue *lvalue,
       rv = FALSE;
     } else {
       ctpl_value_set_float (value, lval / rval);
-      //~ g_debug ("division result: %f", ctpl_value_get_float (value));
     }
   }
   
   return rv;
 }
 
+/*
+ * ctpl_eval_operator_cmp:
+ * @lvalue: Input left operand (may be modified for internal purpose)
+ * @rvalue: Input right operand (may be modified for internal purpose)
+ * @op: The operator, only used for error messages.
+ * @result (out): Result of the operation: 0 if values are equal, < 0 if @lvalue
+ *                is inferior to @rvalue and > 0 if @lvalue is superior to
+ *                @rvalue
+ * @error: a #GError to fill with an eventual error, of %NULL to ignore errors.
+ * 
+ * Compares two values to a strcmp()-like value.
+ * Errors can happen if the types of the operand can't be compared.
+ * 
+ * See also ctpl_eval_operator_sup_inf_eq_neq_supeq_infeq().
+ * 
+ * Returns: %TRUE on success, %FALSE on failure.
+ */
+static gboolean
+ctpl_eval_operator_cmp (CtplValue     *lvalue,
+                        CtplValue     *rvalue,
+                        CtplOperator   op,
+                        gint          *result,
+                        GError       **error)
+{
+  gboolean rv = TRUE;
+  
+  *result = 0;
+  switch (ctpl_value_get_held_type (lvalue)) {
+    case CTPL_VTYPE_ARRAY:
+      if (! CTPL_VALUE_HOLDS_ARRAY (rvalue)) {
+        g_set_error (error, CTPL_EVAL_ERROR, CTPL_EVAL_ERROR_INVALID_OPERAND,
+                     "Invalid operands for operator '%s' (have '%s' and '%s')",
+                     ctpl_operator_to_string (op),
+                     ctpl_value_get_held_type_name (lvalue),
+                     ctpl_value_get_held_type_name (rvalue));
+        rv = FALSE;
+      } else {
+        const GSList *larray;
+        const GSList *rarray;
+        
+        larray = ctpl_value_get_array (lvalue);
+        rarray = ctpl_value_get_array (rvalue);
+        for (; rv && *result == 0 && larray && rarray;
+             larray = larray->next, rarray = rarray->next) {
+          rv = ctpl_eval_operator_cmp (larray->data, rarray->data, op, result,
+                                       error);
+        }
+        if (rv && *result == 0) {
+          if (! larray && rarray) {
+            *result = -1;
+          } else if (larray && ! rarray) {
+            *result = 1;
+          }
+        }
+      }
+      break;
+    
+    case CTPL_VTYPE_INT:
+      if (CTPL_VALUE_HOLDS_INT (rvalue)) {
+        glong lval;
+        glong rval;
+        
+        lval = ctpl_value_get_int (lvalue);
+        rval = ctpl_value_get_int (rvalue);
+        /* no (lval - rval) because of possible over/underflow in subtraction */
+        if (lval < rval) {
+          *result = -1;
+        } else if (lval > rval) {
+          *result = 1;
+        } else {
+          *result = 0;
+        }
+        break;
+      }
+      /* WARNING: conditional break to fall back to floating conversion if one
+       * operand is float */
+    case CTPL_VTYPE_FLOAT:
+      rv = ensure_operands_type (lvalue, rvalue, CTPL_VTYPE_FLOAT,
+                                 ctpl_operator_to_string (op), error);
+      if (rv) {
+        gdouble lval;
+        gdouble rval;
+        
+        lval = ctpl_value_get_float (lvalue);
+        rval = ctpl_value_get_float (rvalue);
+        if (CTPL_MATH_FLOAT_EQ (lval, rval)) {
+          *result = 0;
+        } else if (lval < rval) {
+          *result = -1;
+        } else if (lval > rval) {
+          *result = 1;
+        } else {
+          /* FIXME: should never happen, but what if CTPL_FLOAT_EQ() is more 
+           * precise than < and >? guess it's impossible... */
+          g_return_val_if_reached (FALSE);
+        }
+      }
+      break;
+    
+    case CTPL_VTYPE_STRING:
+      if (CTPL_VALUE_HOLDS_ARRAY (rvalue)) {
+        g_set_error (error, CTPL_EVAL_ERROR, CTPL_EVAL_ERROR_INVALID_OPERAND,
+                     "Invalid operands for operator '%s' (have '%s' and '%s')",
+                     ctpl_operator_to_string (op),
+                     ctpl_value_get_held_type_name (lvalue),
+                     ctpl_value_get_held_type_name (rvalue));
+        rv = FALSE;
+      } else {
+        gchar *tmp;
+        
+        tmp = ctpl_value_to_string (rvalue);
+        *result = strcmp (ctpl_value_get_string (lvalue), tmp);
+        g_free (tmp);
+      }
+      break;
+  }
+  
+  return rv;
+}
+
 /*
  * ctpl_eval_operator_sup_inf_eq_neq_supeq_infeq:
  * @lvalue: Input left operand (may be modified for internal purpose)
@@ -392,6 +511,8 @@ ctpl_eval_operator_div (CtplValue *lvalue,
  * Evaluates superiority, inferiority, equality and derived.
  * Errors can happen if the types of the operand can't be compared.
  * 
+ * See also ctpl_eval_operator_cmp().
+ * 
  * Returns: %TRUE on success, %FALSE on failure.
  */
 static gboolean
@@ -402,93 +523,23 @@ ctpl_eval_operator_sup_inf_eq_neq_supeq_infeq (CtplValue     *lvalue,
                                                GError       **error)
 {
   /* boolean operators expands to integers (1:TRUE or 0:FALSE) */
-  gboolean rv     = TRUE;
-  gboolean result = FALSE;
+  gboolean  rv     = TRUE;
+  gboolean  result = FALSE;
+  gint      r      = 0;
   
-  switch (ctpl_value_get_held_type (lvalue)) {
-    case CTPL_VTYPE_ARRAY:
-      if (ctpl_value_get_held_type (rvalue) != CTPL_VTYPE_ARRAY) {
-        /* fail, can't comapre array with other things */
-        rv = FALSE;
-      } else {
-        const GSList *larray;
-        const GSList *rarray;
-        
-        larray = ctpl_value_get_array (lvalue);
-        rarray = ctpl_value_get_array (rvalue);
-        /*TODO*/
-      }
-      break;
-    
-    case CTPL_VTYPE_INT:
-      if (CTPL_VALUE_HOLDS_INT (rvalue)) {
-        glong lval;
-        glong rval;
-        
-        lval = ctpl_value_get_int (lvalue);
-        rval = ctpl_value_get_int (rvalue);
-        switch (op) {
-          case CTPL_OPERATOR_EQUAL: result = (lval == rval); break;
-          case CTPL_OPERATOR_NEQ:   result = (lval != rval); break;
-          case CTPL_OPERATOR_INF:   result = (lval <  rval); break;
-          case CTPL_OPERATOR_INFEQ: result = (lval <= rval); break;
-          case CTPL_OPERATOR_SUP:   result = (lval >  rval); break;
-          case CTPL_OPERATOR_SUPEQ: result = (lval >= rval); break;
-          default: /* avoid compiler warnings about non-handled cases */ break;
-        }
-        break;
-      }
-      /* WARNING: conditional break to fall back to floating conversion if one
-       * operand is float */
-    case CTPL_VTYPE_FLOAT:
-      rv = ensure_operands_type (lvalue, rvalue, CTPL_VTYPE_FLOAT,
-                                 ctpl_operator_to_string (op), error);
-      if (rv) {
-        gdouble lval;
-        gdouble rval;
-        
-        lval = ctpl_value_get_float (lvalue);
-        rval = ctpl_value_get_float (rvalue);
-        switch (op) {
-          case CTPL_OPERATOR_EQUAL: result = CTPL_MATH_FLOAT_EQ (lval, rval); break;
-          case CTPL_OPERATOR_NEQ:   result = ! CTPL_MATH_FLOAT_EQ (lval, rval); break;
-          case CTPL_OPERATOR_INF:   result = (lval <  rval); break;
-          case CTPL_OPERATOR_INFEQ: result = (lval <= rval); break;
-          case CTPL_OPERATOR_SUP:   result = (lval >  rval); break;
-          case CTPL_OPERATOR_SUPEQ: result = (lval >= rval); break;
-          default: /* avoid compiler warnings about non-handled cases */ break;
-        }
-      }
-      break;
-    
-    case CTPL_VTYPE_STRING:
-      if (CTPL_VALUE_HOLDS_ARRAY (rvalue)) {
-        g_set_error (error, CTPL_EVAL_ERROR, CTPL_EVAL_ERROR_INVALID_OPERAND,
-                     "Invalid operands for operator '%s' (have '%s' and '%s')",
-                     ctpl_operator_to_string (op),
-                     ctpl_value_get_held_type_name (lvalue),
-                     ctpl_value_get_held_type_name (rvalue));
-        rv = FALSE;
-      } else {
-        gchar  *tmp = NULL;
-        gint    strcmp_result;
-        
-        tmp = ctpl_value_to_string (rvalue);
-        strcmp_result = strcmp (ctpl_value_get_string (lvalue), tmp);
-        switch (op) {
-          case CTPL_OPERATOR_EQUAL: result = (strcmp_result == 0); break;
-          case CTPL_OPERATOR_NEQ:   result = (strcmp_result != 0); break;
-          case CTPL_OPERATOR_INF:   result = (strcmp_result <  0); break;
-          case CTPL_OPERATOR_INFEQ: result = (strcmp_result <= 0); break;
-          case CTPL_OPERATOR_SUP:   result = (strcmp_result >  0); break;
-          case CTPL_OPERATOR_SUPEQ: result = (strcmp_result >= 0); break;
-          default: /* avoid compiler warnings about non-handled cases */ break;
-        }
-        g_free (tmp);
-      }
-      break;
+  rv = ctpl_eval_operator_cmp (lvalue, rvalue, op, &r, error);
+  if (rv) {
+    switch (op) {
+      case CTPL_OPERATOR_EQUAL: result = (r == 0); break;
+      case CTPL_OPERATOR_NEQ:   result = (r != 0); break;
+      case CTPL_OPERATOR_INF:   result = (r <  0); break;
+      case CTPL_OPERATOR_INFEQ: result = (r <= 0); break;
+      case CTPL_OPERATOR_SUP:   result = (r >  0); break;
+      case CTPL_OPERATOR_SUPEQ: result = (r >= 0); break;
+      default: /* avoid compiler warnings about non-handled cases */ break;
+    }
   }
-  ctpl_value_set_int (value, result ? 1 : 0);
+  ctpl_value_set_int (value, result ? 1L : 0L);
   
   return rv;
 }
@@ -723,12 +774,6 @@ ctpl_eval_value (const CtplTokenExpr  *expr,
     }
     g_free (value_str);
   }
-  //~ {
-    //~ gchar *dump;
-    //~ dump = ctpl_value_to_string (value);
-    //~ g_debug ("result: %s", dump);
-    //~ g_free (dump);
-  //~ }
   
   return rv;
 }
