@@ -1,6 +1,6 @@
 /* 
  * 
- * Copyright (C) 2009-2010 Colomban Wendling <ban@herbesfolles.org>
+ * Copyright (C) 2009-2011 Colomban Wendling <ban@herbesfolles.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -112,6 +112,36 @@ ctpl_lexer_error_quark (void)
 }
 
 
+/* Reads a statement end (the "}" part) */
+static gboolean
+ctpl_lexer_read_stmt_end (CtplInputStream  *stream,
+                          const gchar      *stmt_name,
+                          GError          **error)
+{
+  gboolean success = FALSE;
+  
+  if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
+    GError *err = NULL;
+    gint    c;
+    
+    c = ctpl_input_stream_get_c (stream, &err);
+    if (err) {
+      /* I/O error */
+      g_propagate_error (error, err);
+    } else if (c != CTPL_END_CHAR) {
+      /* trash before the end, fail */
+      ctpl_input_stream_set_error (stream, error, CTPL_LEXER_ERROR,
+                                   CTPL_LEXER_ERROR_SYNTAX_ERROR,
+                                   "Unexpected character '%c' before end of "
+                                   "'%s' statement", c, stmt_name);
+    } else {
+      success = TRUE;
+    }
+  }
+  
+  return success;
+}
+
 /* reads the data part of a if, aka the expression (e.g. " a > b" in "if a > b")
  * Return a new token or %NULL on error */
 static CtplToken *
@@ -122,52 +152,35 @@ ctpl_lexer_read_token_tpl_if (CtplInputStream *stream,
   CtplToken      *token = NULL;
   CtplTokenExpr  *expr;
   
-  //~ g_debug ("if?");
-  if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
-    expr = ctpl_lexer_expr_lex_full (stream, FALSE, error);
-    if (expr) {
-      GError *err = NULL;
-      gint    c;
+  expr = ctpl_lexer_expr_lex_full (stream, FALSE, error);
+  if (expr) {
+    if (ctpl_lexer_read_stmt_end (stream, "if", error)) {
+      GError     *err = NULL;
+      CtplToken  *if_token;
+      CtplToken  *else_token = NULL;
+      LexerState  substate = *state;
       
-      c = ctpl_input_stream_get_c (stream, &err);
-      if (err) {
-        /* I/O error */
-      } else if (c != CTPL_END_CHAR) {
-        /* there is trash before the end, then fail */
-        ctpl_input_stream_set_error (stream, error, CTPL_LEXER_ERROR,
-                                     CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                     "Unexpected character '%c' before end of "
-                                     "'if' statement", c);
-      } else {
-        CtplToken  *if_token;
-        CtplToken  *else_token = NULL;
-        LexerState  substate = *state;
-        
-        //~ g_debug ("if token: `if %s`", expr);
-        substate.block_depth ++;
-        substate.last_statement_type_if = S_IF;
-        if_token = ctpl_lexer_lex_internal (stream, &substate, &err);
+      substate.block_depth ++;
+      substate.last_statement_type_if = S_IF;
+      if_token = ctpl_lexer_lex_internal (stream, &substate, &err);
+      if (! err) {
+        if (substate.last_statement_type_if == S_ELSE) {
+          else_token = ctpl_lexer_lex_internal (stream, &substate, &err);
+        }
+        if (! err /* don't override errors */ &&
+            state->block_depth != substate.block_depth) {
+          /* if a block was not closed, fail */
+          ctpl_input_stream_set_error (stream, &err, CTPL_LEXER_ERROR,
+                                       CTPL_LEXER_ERROR_SYNTAX_ERROR,
+                                       "Unclosed 'if/else' block");
+        }
         if (! err) {
-          if (substate.last_statement_type_if == S_ELSE) {
-            //~ g_debug ("have else");
-            else_token = ctpl_lexer_lex_internal (stream, &substate, &err);
-          }
-          if (! err /* don't override errors */ &&
-              state->block_depth != substate.block_depth) {
-            /* if a block was not closed, fail */
-            ctpl_input_stream_set_error (stream, &err, CTPL_LEXER_ERROR,
-                                         CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                         "Unclosed 'if/else' block");
-          }
-          if (! err) {
-            token = ctpl_token_new_if (expr, if_token, else_token);
-            /* set expr to NULL not to free it since it is now used */
-            expr = NULL;
-          }
+          token = ctpl_token_new_if (expr, if_token, else_token);
+          /* set expr to NULL not to free it since it is now used */
+          expr = NULL;
+        } else {
+          g_propagate_error (error, err);
         }
-      }
-      if (err) {
-        g_propagate_error (error, err);
       }
     }
     ctpl_token_expr_free (expr);
@@ -185,7 +198,6 @@ ctpl_lexer_read_token_tpl_for (CtplInputStream *stream,
 {
   CtplToken  *token = NULL;
   
-  //~ g_debug ("for?");
   if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
     gchar *iter_name;
     
@@ -198,7 +210,6 @@ ctpl_lexer_read_token_tpl_for (CtplInputStream *stream,
                                    CTPL_LEXER_ERROR_SYNTAX_ERROR,
                                    "No iterator identifier for 'for' statement");
     } else {
-      //~ g_debug ("for: iter is '%s'", iter_name);
       if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
         gchar *keyword_in;
         
@@ -212,51 +223,31 @@ ctpl_lexer_read_token_tpl_for (CtplInputStream *stream,
                                        "Missing 'in' keyword after iterator name "
                                        "of 'for' statement");
         } else {
-          if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
-            CtplTokenExpr *array_expr;
-            
-            array_expr = ctpl_lexer_expr_lex_full (stream, FALSE, error);
-            if (! array_expr) {
-            } else {
-              if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
-                gint    c;
-                GError *err = NULL;
-                
-                c = ctpl_input_stream_get_c (stream, &err);
-                if (err) {
-                  /* I/O error */
-                  g_propagate_error (error, err);
-                } else if (c != CTPL_END_CHAR) {
-                  /* trash before the end, fail */
-                  ctpl_input_stream_set_error (stream, error, CTPL_LEXER_ERROR,
+          CtplTokenExpr *array_expr;
+          
+          array_expr = ctpl_lexer_expr_lex_full (stream, FALSE, error);
+          if (array_expr) {
+            if (ctpl_lexer_read_stmt_end (stream, "for", error)) {
+              GError     *err = NULL;
+              CtplToken  *for_children;
+              LexerState  substate = *state;
+              
+              substate.block_depth ++;
+              for_children = ctpl_lexer_lex_internal (stream, &substate, &err);
+              if (! err) {
+                if (state->block_depth != substate.block_depth) {
+                  ctpl_input_stream_set_error (stream, &err, CTPL_LEXER_ERROR,
                                                CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                               "Unexpected character '%c' "
-                                               "before end of 'for' statement",
-                                               c);
+                                               "Unclosed 'for' block");
                 } else {
-                  CtplToken  *for_children;
-                  LexerState  substate = *state;
-                  
-                  //~ g_debug ("for token: `for %s in %s`", iter_name, array_name);
-                  substate.block_depth ++;
-                  for_children = ctpl_lexer_lex_internal (stream, &substate, &err);
-                  //~ g_debug ("for child: %d", (void *)for_children);
-                  if (! err) {
-                    if (state->block_depth != substate.block_depth) {
-                      ctpl_input_stream_set_error (stream, &err, CTPL_LEXER_ERROR,
-                                                   CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                                   "Unclosed 'for' block");
-                    } else {
-                      token = ctpl_token_new_for_expr (array_expr, iter_name,
-                                                       for_children);
-                      array_expr = NULL; /* avoid freeing expression */
-                    }
-                  }
-                  if (err) {
-                    g_propagate_error (error, err);
-                  }
+                  token = ctpl_token_new_for (array_expr, iter_name,
+                                              for_children);
+                  array_expr = NULL; /* avoid freeing expression */
                 }
               }
+              if (err) {
+                g_propagate_error (error, err);
+              }
             }
             ctpl_token_expr_free (array_expr);
           }
@@ -271,88 +262,48 @@ ctpl_lexer_read_token_tpl_for (CtplInputStream *stream,
 }
 
 /* reads an end block end (} of a {end} block)
- * Returns: %TRUE on success, %FALSE otherwise. */
-static gboolean
+ * Always returns %NULL to stop lexing pass or notify an error. */
+static CtplToken *
 ctpl_lexer_read_token_tpl_end (CtplInputStream *stream,
                                LexerState      *state,
                                GError         **error)
 {
-  gboolean  rv = FALSE;
-  
-  //~ g_debug ("end?");
-  if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
-    gint    c;
-    GError *err = NULL;
-    
-    c = ctpl_input_stream_get_c (stream, &err);
-    if (err) {
-      /* I/O error */
-      g_propagate_error (error, err);
-    } else if (c != CTPL_END_CHAR) {
-      /* fail, missing } at the end */
+  if (ctpl_lexer_read_stmt_end (stream, "end", error)) {
+    state->block_depth --;
+    if (state->block_depth < 0) {
+      /* a non-opened block was closed, fail */
       ctpl_input_stream_set_error (stream, error, CTPL_LEXER_ERROR,
                                    CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                   "Unexpected character '%c' before end of "
-                                   "'end' statement", c);
+                                   "Unmatching 'end' statement (needs a 'if' "
+                                   "or 'for' before)");
     } else {
-      //~ g_debug ("block end");
-      state->block_depth --;
-      if (state->block_depth < 0) {
-        /* a non-opened block was closed, fail */
-        ctpl_input_stream_set_error (stream, error, CTPL_LEXER_ERROR,
-                                     CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                     "Unmatching 'end' statement (needs a 'if' "
-                                     "or 'for' before)");
-      } else {
-        state->last_statement_type_if = S_END;
-        rv = TRUE;
-      }
+      state->last_statement_type_if = S_END;
     }
   }
   
-  return rv;
+  return NULL;
 }
 
 /* reads an else block end (} of a {else} block)
- * Returns: %TRUE on success, %FALSE otherwise. */
-static gboolean
+ * Always returns %NULL to stop lexing pass or notify an error. */
+static CtplToken *
 ctpl_lexer_read_token_tpl_else (CtplInputStream *stream,
                                 LexerState      *state,
                                 GError         **error)
 {
-  gboolean  rv = FALSE;
-  
-  //~ g_debug ("else?");
-  if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
-    GError *err = NULL;
-    gint    c;
-    
-    c = ctpl_input_stream_get_c (stream, &err);
-    if (err) {
-      /* I/O error */
-      g_propagate_error (error, err);
-    } else if (c != CTPL_END_CHAR) {
-      /* fail, missing } at the end */
+  if (ctpl_lexer_read_stmt_end (stream, "else", error)) {
+    if (state->last_statement_type_if != S_IF) {
+      /* else but no opened if, fail */
       ctpl_input_stream_set_error (stream, error, CTPL_LEXER_ERROR,
                                    CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                   "Unexpected character '%c' before end of "
-                                   "'else' statement", c);
+                                   "Unmatching 'else' statement (needs an "
+                                   "'if' before)");
     } else {
-      //~ g_debug ("else statement");
-      if (state->last_statement_type_if != S_IF) {
-        /* else but no opened if, fail */
-        ctpl_input_stream_set_error (stream, error, CTPL_LEXER_ERROR,
-                                     CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                     "Unmatching 'else' statement (needs an "
-                                     "'if' before)");
-      } else {
-        state->last_statement_type_if = S_ELSE;
-        rv = TRUE;
-      }
+      state->last_statement_type_if = S_ELSE;
     }
   }
   
-  return rv;
+  return NULL;
 }
 
 /* Reads an expression token (:BLANKCHARS:?:EXPRCHARS::BLANKCHARS:?}, without
@@ -366,28 +317,12 @@ ctpl_lexer_read_token_tpl_expr (CtplInputStream *stream,
   CtplTokenExpr  *expr;
   
   (void)state; /* we don't use the state, silent compilers */
-  if (ctpl_input_stream_skip_blank (stream, error) >= 0) {
-    expr = ctpl_lexer_expr_lex_full (stream, FALSE, error);
-    if (expr) {
-      GError *err = NULL;
-      gint    c;
-      
-      c = ctpl_input_stream_get_c (stream, &err);
-      if (err) {
-        /* I/O error */
-        g_propagate_error (error, err);
-      } else if (c != CTPL_END_CHAR) {
-        /* trash before the end, fail */
-        ctpl_input_stream_set_error (stream, error, CTPL_LEXER_ERROR,
-                                     CTPL_LEXER_ERROR_SYNTAX_ERROR,
-                                     "Unexpected character '%c' before end of "
-                                     "statement", c);
-      } else {
-        token = ctpl_token_new_expr (expr);
-      }
-      if (! token) {
-        ctpl_token_expr_free (expr);
-      }
+  expr = ctpl_lexer_expr_lex_full (stream, FALSE, error);
+  if (expr) {
+    if (ctpl_lexer_read_stmt_end (stream, "expression", error)) {
+      token = ctpl_token_new_expr (expr);
+    } else {
+      ctpl_token_expr_free (expr);
     }
   }
   
@@ -424,36 +359,29 @@ ctpl_lexer_read_token_tpl (CtplInputStream *stream,
       first_word = ctpl_input_stream_peek_symbol_full (stream, 5,
                                                        &first_word_len, error);
       if (first_word) {
-        /* FIXME: there is lot of duplicates */
-        if (strcmp (first_word, "if") == 0) {
-          /* an if condition:
-           * if expr */
-          ctpl_input_stream_skip (stream, first_word_len, NULL);
-          token = ctpl_lexer_read_token_tpl_if (stream, state, error);
-        } else if (strcmp (first_word, "for") == 0) {
-          /* a for loop:
-           * for iter in array */
-          ctpl_input_stream_skip (stream, first_word_len, NULL);
-          token = ctpl_lexer_read_token_tpl_for (stream, state, error);
-        } else if (strcmp (first_word, "end") == 0) {
-          /* a block end:
-           * {end} */
-          ctpl_input_stream_skip (stream, first_word_len, NULL);
-          ctpl_lexer_read_token_tpl_end (stream, state, error);
-          /* here we will return NULL, which is fine as it simply stops token
-           * reading, and as we use nested lexing calls for nested blocks */
-        } else if (strcmp (first_word, "else") == 0) {
-          /* an else statement:
-           * {else} */
-          ctpl_input_stream_skip (stream, first_word_len, NULL);
-          ctpl_lexer_read_token_tpl_else (stream, state, error);
-          /* return NULL, see above */
+        /* tries to handle @keyword, returns whether it has been handled.
+         * @handler must return a token on success, or %NULL to stop lexing
+         * (not necessarily an error, may end a nested pass) */
+        #define HANDLE_KEYWORD(keyword, handler)                               \
+          ((strcmp (first_word, keyword) == 0)                                 \
+           ? (ctpl_input_stream_skip (stream, first_word_len, NULL),           \
+              (token = handler (stream, state, error)),                        \
+              TRUE)                                                            \
+           : FALSE)
+        
+        if        (HANDLE_KEYWORD ("if",    ctpl_lexer_read_token_tpl_if)) {
+        } else if (HANDLE_KEYWORD ("for",   ctpl_lexer_read_token_tpl_for)) {
+        } else if (HANDLE_KEYWORD ("end",   ctpl_lexer_read_token_tpl_end)) {
+        } else if (HANDLE_KEYWORD ("else",  ctpl_lexer_read_token_tpl_else)) {
         } else {
-          /* an expression, or nothing valid */
+          /* if nothing matched, it's an expression or nothing valid */
           token = ctpl_lexer_read_token_tpl_expr (stream, state, error);
         }
+        
+        #undef HANDLE_KEYWORD
+        
+        g_free (first_word);
       }
-      g_free (first_word);
     }
   }
   
@@ -531,10 +459,8 @@ ctpl_lexer_read_token (CtplInputStream *stream,
   if (err) {
     g_propagate_error (error, err);
   } else {
-    //~ g_debug ("Will read a token (starts with %c)", c);
     switch (c) {
       case CTPL_START_CHAR:
-        //~ g_debug ("start of a template recognized syntax");
         token = ctpl_lexer_read_token_tpl (stream, state, error);
         break;
       
